@@ -13,7 +13,8 @@ from .settings import FILE_SKIP_COMMENTS
 
 CIMPORT_IDENTIFIERS = ("cimport ", "cimport*", "from.cimport")
 IMPORT_START_IDENTIFIERS = ("from ", "from.import", "import ", "import*") + CIMPORT_IDENTIFIERS
-COMMENT_INDICATORS = ('"""', "'''", "'", '"', "#")
+DOCSTRING_INDICATORS = ('"""', "'''")
+COMMENT_INDICATORS = DOCSTRING_INDICATORS + ("'", '"', "#")
 CODE_SORT_COMMENTS = (
     "# isort: list",
     "# isort: dict",
@@ -29,6 +30,7 @@ def process(
     input_stream: TextIO,
     output_stream: TextIO,
     extension: str = "py",
+    raise_on_skip: bool = True,
     config: Config = DEFAULT_CONFIG,
 ) -> bool:
     """Parses stream identifying sections of contiguous imports and sorting them
@@ -53,14 +55,15 @@ def process(
     next_import_section: str = ""
     next_cimports: bool = False
     in_quote: str = ""
+    was_in_quote: bool = False
     first_comment_index_start: int = -1
     first_comment_index_end: int = -1
     contains_imports: bool = False
     in_top_comment: bool = False
     first_import_section: bool = True
-    section_comments = [f"# {heading}" for heading in config.import_headings.values()]
     indent: str = ""
     isort_off: bool = False
+    skip_file: bool = False
     code_sorting: Union[bool, str] = False
     code_sorting_section: str = ""
     code_sorting_indent: str = ""
@@ -68,6 +71,7 @@ def process(
     made_changes: bool = False
     stripped_line: str = ""
     end_of_file: bool = False
+    verbose_output: List[str] = []
 
     if config.float_to_top:
         new_input = ""
@@ -84,9 +88,14 @@ def process(
                 if line == "# isort: off\n":
                     isort_off = True
                 if current:
+                    if add_imports:
+                        add_line_separator = line_separator or "\n"
+                        current += add_line_separator + add_line_separator.join(add_imports)
+                        add_imports = []
                     parsed = parse.file_contents(current, config=config)
+                    verbose_output += parsed.verbose_output
                     extra_space = ""
-                    while current[-1] == "\n":
+                    while current and current[-1] == "\n":
                         extra_space += "\n"
                         current = current[:-1]
                     extra_space = extra_space.replace("\n", "", 1)
@@ -120,17 +129,22 @@ def process(
                 line_separator = "\n"
 
             if code_sorting and code_sorting_section:
-                output_stream.write(
-                    textwrap.indent(
-                        isort.literal.assignment(
-                            code_sorting_section,
-                            str(code_sorting),
-                            extension,
-                            config=_indented_config(config, indent),
-                        ),
-                        code_sorting_indent,
-                    )
+                sorted_code = textwrap.indent(
+                    isort.literal.assignment(
+                        code_sorting_section,
+                        str(code_sorting),
+                        extension,
+                        config=_indented_config(config, indent),
+                    ),
+                    code_sorting_indent,
                 )
+                made_changes = made_changes or _has_changed(
+                    before=code_sorting_section,
+                    after=sorted_code,
+                    line_separator=line_separator,
+                    ignore_whitespace=config.ignore_whitespace,
+                )
+                output_stream.write(sorted_code)
         else:
             stripped_line = line.strip()
             if stripped_line and not line_separator:
@@ -138,22 +152,42 @@ def process(
 
             for file_skip_comment in FILE_SKIP_COMMENTS:
                 if file_skip_comment in line:
-                    raise FileSkipComment("Passed in content")
+                    if raise_on_skip:
+                        raise FileSkipComment("Passed in content")
+                    isort_off = True
+                    skip_file = True
 
-            if not in_quote and stripped_line == "# isort: off":
-                isort_off = True
+            if not in_quote:
+                if stripped_line == "# isort: off":
+                    isort_off = True
+                elif stripped_line.startswith("# isort: dont-add-imports"):
+                    add_imports = []
+                elif stripped_line.startswith("# isort: dont-add-import:"):
+                    import_not_to_add = stripped_line.split("# isort: dont-add-import:", 1)[
+                        1
+                    ].strip()
+                    add_imports = [
+                        import_to_add
+                        for import_to_add in add_imports
+                        if not import_to_add == import_not_to_add
+                    ]
 
             if (
                 (index == 0 or (index in (1, 2) and not contains_imports))
                 and stripped_line.startswith("#")
-                and stripped_line not in section_comments
+                and stripped_line not in config.section_comments
+                and stripped_line not in CODE_SORT_COMMENTS
             ):
                 in_top_comment = True
-            elif in_top_comment:
-                if not line.startswith("#") or stripped_line in section_comments:
-                    in_top_comment = False
-                    first_comment_index_end = index - 1
+            elif in_top_comment and (
+                not line.startswith("#")
+                or stripped_line in config.section_comments
+                or stripped_line in CODE_SORT_COMMENTS
+            ):
+                in_top_comment = False
+                first_comment_index_end = index - 1
 
+            was_in_quote = bool(in_quote)
             if (not stripped_line.startswith("#") or in_quote) and '"' in line or "'" in line:
                 char_index = 0
                 if first_comment_index_start == -1 and (
@@ -179,10 +213,10 @@ def process(
                         break
                     char_index += 1
 
-            not_imports = bool(in_quote) or in_top_comment or isort_off
-            if not (in_quote or in_top_comment):
+            not_imports = bool(in_quote) or was_in_quote or in_top_comment or isort_off
+            if not (in_quote or was_in_quote or in_top_comment):
                 if isort_off:
-                    if stripped_line == "# isort: on":
+                    if not skip_file and stripped_line == "# isort: on":
                         isort_off = False
                 elif stripped_line.endswith("# isort: split"):
                     not_imports = True
@@ -192,17 +226,22 @@ def process(
                     not_imports = True
                 elif code_sorting:
                     if not stripped_line:
-                        output_stream.write(
-                            textwrap.indent(
-                                isort.literal.assignment(
-                                    code_sorting_section,
-                                    str(code_sorting),
-                                    extension,
-                                    config=_indented_config(config, indent),
-                                ),
-                                code_sorting_indent,
-                            )
+                        sorted_code = textwrap.indent(
+                            isort.literal.assignment(
+                                code_sorting_section,
+                                str(code_sorting),
+                                extension,
+                                config=_indented_config(config, indent),
+                            ),
+                            code_sorting_indent,
                         )
+                        made_changes = made_changes or _has_changed(
+                            before=code_sorting_section,
+                            after=sorted_code,
+                            line_separator=line_separator,
+                            ignore_whitespace=config.ignore_whitespace,
+                        )
+                        output_stream.write(sorted_code)
                         not_imports = True
                         code_sorting = False
                         code_sorting_section = ""
@@ -210,8 +249,13 @@ def process(
                     else:
                         code_sorting_section += line
                         line = ""
-                elif stripped_line in config.section_comments and not import_section:
-                    import_section += line
+                elif stripped_line in config.section_comments:
+                    if import_section and not contains_imports:
+                        output_stream.write(import_section)
+                        import_section = line
+                        not_imports = False
+                    else:
+                        import_section += line
                     indent = line[: -len(line.lstrip())]
                 elif not (stripped_line or contains_imports):
                     not_imports = True
@@ -224,8 +268,6 @@ def process(
                 ):
                     import_section += line
                 elif stripped_line.startswith(IMPORT_START_IDENTIFIERS):
-                    contains_imports = True
-
                     new_indent = line[: -len(line.lstrip())]
                     import_statement = line
                     stripped_line = line.strip().split("#")[0]
@@ -243,28 +285,47 @@ def process(
                                 stripped_line = line.strip().split("#")[0]
                                 import_statement += line
 
-                    cimport_statement: bool = False
                     if (
-                        import_statement.lstrip().startswith(CIMPORT_IDENTIFIERS)
-                        or " cimport " in import_statement
-                        or " cimport*" in import_statement
-                        or " cimport(" in import_statement
-                        or ".cimport" in import_statement
+                        import_statement.lstrip().startswith("from")
+                        and "import" not in import_statement
                     ):
-                        cimport_statement = True
+                        line = import_statement
+                        not_imports = True
+                    else:
+                        did_contain_imports = contains_imports
+                        contains_imports = True
 
-                    if cimport_statement != cimports or (new_indent != indent and import_section):
-                        if import_section:
-                            next_cimports = cimport_statement
-                            next_import_section = import_statement
-                            import_statement = ""
-                            not_imports = True
-                            line = ""
+                        cimport_statement: bool = False
+                        if (
+                            import_statement.lstrip().startswith(CIMPORT_IDENTIFIERS)
+                            or " cimport " in import_statement
+                            or " cimport*" in import_statement
+                            or " cimport(" in import_statement
+                            or ".cimport" in import_statement
+                        ):
+                            cimport_statement = True
+
+                        if cimport_statement != cimports or (
+                            new_indent != indent
+                            and import_section
+                            and (not did_contain_imports or len(new_indent) < len(indent))
+                        ):
+                            indent = new_indent
+                            if import_section:
+                                next_cimports = cimport_statement
+                                next_import_section = import_statement
+                                import_statement = ""
+                                not_imports = True
+                                line = ""
+                            else:
+                                cimports = cimport_statement
                         else:
-                            cimports = cimport_statement
-
-                    indent = new_indent
-                    import_section += import_statement
+                            if new_indent != indent:
+                                if import_section and did_contain_imports:
+                                    import_statement = indent + import_statement.lstrip()
+                                else:
+                                    indent = new_indent
+                        import_section += import_statement
                 else:
                     not_imports = True
 
@@ -275,13 +336,15 @@ def process(
                 and (stripped_line or end_of_file)
                 and not config.append_only
                 and not in_top_comment
-                and not in_quote
+                and not was_in_quote
                 and not import_section
                 and not line.lstrip().startswith(COMMENT_INDICATORS)
+                and not (line.rstrip().endswith(DOCSTRING_INDICATORS) and "=" not in line)
             ):
-                import_section = line_separator.join(add_imports) + line_separator
+                add_line_separator = line_separator or "\n"
+                import_section = add_line_separator.join(add_imports) + add_line_separator
                 if end_of_file and index != 0:
-                    output_stream.write(line_separator)
+                    output_stream.write(add_line_separator)
                 contains_imports = True
                 add_imports = []
 
@@ -302,6 +365,7 @@ def process(
                     raw_import_section += line
                 if not contains_imports:
                     output_stream.write(import_section)
+
                 else:
                     leading_whitespace = import_section[: -len(import_section.lstrip())]
                     trailing_whitespace = import_section[len(import_section.rstrip()) :]
@@ -317,8 +381,11 @@ def process(
                             line[len(indent) :] for line in import_section.splitlines(keepends=True)
                         )
 
+                    parsed_content = parse.file_contents(import_section, config=config)
+                    verbose_output += parsed_content.verbose_output
+
                     sorted_import_section = output.sorted_imports(
-                        parse.file_contents(import_section, config=config),
+                        parsed_content,
                         _indented_config(config, indent),
                         extension,
                         import_type="cimport" if cimports else "import",
@@ -337,7 +404,6 @@ def process(
                             line_separator=line_separator,
                             ignore_whitespace=config.ignore_whitespace,
                         )
-
                         output_stream.write(sorted_import_section)
                         if not line and not indent and next_import_section:
                             output_stream.write(line_separator)
@@ -358,10 +424,33 @@ def process(
                 output_stream.write(line)
                 not_imports = False
 
+            if stripped_line and not in_quote and not import_section and not next_import_section:
+                if stripped_line == "yield":
+                    while not stripped_line or stripped_line == "yield":
+                        new_line = input_stream.readline()
+                        if not new_line:
+                            break
+
+                        output_stream.write(new_line)
+                        stripped_line = new_line.strip().split("#")[0]
+
+                if stripped_line.startswith("raise") or stripped_line.startswith("yield"):
+                    while stripped_line.endswith("\\"):
+                        new_line = input_stream.readline()
+                        if not new_line:
+                            break
+
+                        output_stream.write(new_line)
+                        stripped_line = new_line.strip().split("#")[0]
+
+    if made_changes and config.only_modified:
+        for output_str in verbose_output:
+            print(output_str)
+
     return made_changes
 
 
-def _indented_config(config: Config, indent: str):
+def _indented_config(config: Config, indent: str) -> Config:
     if not indent:
         return config
 
@@ -370,6 +459,7 @@ def _indented_config(config: Config, indent: str):
         line_length=max(config.line_length - len(indent), 0),
         wrap_length=max(config.wrap_length - len(indent), 0),
         lines_after_imports=1,
+        import_headings=config.import_headings if config.indented_import_headings else {},
     )
 
 
@@ -379,5 +469,4 @@ def _has_changed(before: str, after: str, line_separator: str, ignore_whitespace
             remove_whitespace(before, line_separator=line_separator).strip()
             != remove_whitespace(after, line_separator=line_separator).strip()
         )
-    else:
-        return before.strip() != after.strip()
+    return before.strip() != after.strip()
