@@ -1,7 +1,7 @@
 import copy
 import itertools
 from functools import partial
-from typing import Iterable, List, Set, Tuple
+from typing import Any, Iterable, List, Optional, Set, Tuple, Type
 
 from isort.format import format_simplified
 
@@ -49,7 +49,8 @@ def sorted_imports(
     for section in sections:
         straight_modules = parsed.imports[section]["straight"]
         if not config.only_sections:
-            straight_modules = sorting.naturally(
+            straight_modules = sorting.sort(
+                config,
                 straight_modules,
                 key=lambda key: sorting.module_key(
                     key, config, section_name=section, straight_import=True
@@ -59,7 +60,8 @@ def sorted_imports(
 
         from_modules = parsed.imports[section]["from"]
         if not config.only_sections:
-            from_modules = sorting.naturally(
+            from_modules = sorting.sort(
+                config,
                 from_modules,
                 key=lambda key: sorting.module_key(key, config, section_name=section),
                 reverse=config.reverse_sort,
@@ -105,9 +107,11 @@ def sorted_imports(
                 else:
                     new_section_output.append(line)
             # only_sections options is not imposed if force_sort_within_sections is True
-            new_section_output = sorting.naturally(
+            new_section_output = sorting.sort(
+                config,
                 new_section_output,
                 key=partial(sorting.section_key, config=config),
+                reverse=config.reverse_sort,
             )
 
             # uncollapse comments
@@ -131,7 +135,7 @@ def sorted_imports(
                 if config.dedup_headings:
                     seen_headings.add(section_title)
                 section_comment = f"# {section_title}"
-                if section_comment not in parsed.lines_without_imports[0:1]:
+                if section_comment not in parsed.lines_without_imports[0:1]:  # pragma: no branch
                     section_output.insert(0, section_comment)
 
             if pending_lines_before or not no_lines_before:
@@ -172,7 +176,7 @@ def sorted_imports(
             next_construct = ""
             tail = formatted_output[imports_tail:]
 
-            for index, line in enumerate(tail):
+            for index, line in enumerate(tail):  # pragma: no branch
                 should_skip, in_quote, *_ = parse.skip_line(
                     line,
                     in_quote="",
@@ -189,7 +193,7 @@ def sorted_imports(
                         continue
                     next_construct = line
                     break
-                if in_quote:
+                if in_quote:  # pragma: no branch
                     next_construct = line
                     break
 
@@ -237,7 +241,8 @@ def _with_from_imports(
             not config.no_inline_sort
             or (config.force_single_line and module not in config.single_line_exclusions)
         ) and not config.only_sections:
-            from_imports = sorting.naturally(
+            from_imports = sorting.sort(
+                config,
                 from_imports,
                 key=lambda key: sorting.module_key(
                     key,
@@ -265,7 +270,7 @@ def _with_from_imports(
             if not config.no_inline_sort:
                 for as_import in as_imports:
                     if not config.only_sections:
-                        as_imports[as_import] = sorting.naturally(as_imports[as_import])
+                        as_imports[as_import] = sorting.sort(config, as_imports[as_import])
             for from_import in copy.copy(from_imports):
                 if from_import in as_imports:
                     idx = from_imports.index(from_import)
@@ -336,7 +341,7 @@ def _with_from_imports(
                                     removed=config.ignore_comments,
                                     comment_prefix=config.comment_prefix,
                                 )
-                                for as_import in sorting.naturally(as_imports[from_import])
+                                for as_import in sorting.sort(config, as_imports[from_import])
                             )
 
                         else:
@@ -359,7 +364,7 @@ def _with_from_imports(
                     from_import = from_imports.pop(0)
 
                     if not config.only_sections:
-                        as_imports[from_import] = sorting.naturally(as_imports[from_import])
+                        as_imports[from_import] = sorting.sort(config, as_imports[from_import])
                     from_comments = (
                         parsed.categorized_comments["straight"].get(f"{module}.{from_import}") or []
                     )
@@ -415,14 +420,13 @@ def _with_from_imports(
                 if "*" in from_imports:
                     output.append(
                         with_comments(
-                            _with_star_comments(parsed, module, list(comments or ())),
+                            _with_star_comments(parsed, module, []),
                             f"{import_start}*",
                             removed=config.ignore_comments,
                             comment_prefix=config.comment_prefix,
                         )
                     )
                     from_imports.remove("*")
-                    comments = None
 
                 for from_import in copy.copy(from_imports):
                     comment = (
@@ -607,19 +611,23 @@ def _normalize_empty_lines(lines: List[str]) -> List[str]:
 
 
 class _LineWithComments(str):
-    def __new__(cls, value, comments):
-        instance = super().__new__(cls, value)  # type: ignore
+    comments: List[str]
+
+    def __new__(
+        cls: Type["_LineWithComments"], value: Any, comments: List[str]
+    ) -> "_LineWithComments":
+        instance = super().__new__(cls, value)
         instance.comments = comments
         return instance
 
 
-def _ensure_newline_before_comment(output):
+def _ensure_newline_before_comment(output: List[str]) -> List[str]:
     new_output: List[str] = []
 
-    def is_comment(line):
-        return line and line.startswith("#")
+    def is_comment(line: Optional[str]) -> bool:
+        return line.startswith("#") if line else False
 
-    for line, prev_line in zip(output, [None] + output):
+    for line, prev_line in zip(output, [None] + output):  # type: ignore
         if is_comment(line) and prev_line != "" and not is_comment(prev_line):
             new_output.append("")
         new_output.append(line)
