@@ -12,16 +12,30 @@ import subprocess  # nosec: Needed for gitignore support.
 import sys
 from functools import lru_cache
 from pathlib import Path
-from typing import Any, Callable, Dict, FrozenSet, Iterable, List, Optional, Pattern, Set, Tuple
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    FrozenSet,
+    Iterable,
+    List,
+    Optional,
+    Pattern,
+    Set,
+    Tuple,
+    Type,
+    Union,
+)
 from warnings import warn
 
-from . import stdlibs
+from . import sorting, stdlibs
 from ._future import dataclass, field
-import toml
 from .exceptions import (
     FormattingPluginDoesNotExist,
     InvalidSettingsPath,
     ProfileDoesNotExist,
+    SortingFunctionDoesNotExist,
     UnsupportedSettings,
 )
 from .profiles import profiles
@@ -30,8 +44,14 @@ from .sections import FIRSTPARTY, FUTURE, LOCALFOLDER, STDLIB, THIRDPARTY
 from .wrap_modes import WrapModes
 from .wrap_modes import from_string as wrap_mode_from_string
 
+if TYPE_CHECKING:
+    toml: Any
+else:
+    import toml
+
 _SHEBANG_RE = re.compile(br"^#!.*\bpython[23w]?\b")
-SUPPORTED_EXTENSIONS = frozenset({"py", "pyi", "pyx", "pxd"})
+CYTHON_EXTENSIONS = frozenset({"pyx", "pxd"})
+SUPPORTED_EXTENSIONS = frozenset({"py", "pyi", *CYTHON_EXTENSIONS})
 BLOCKED_EXTENSIONS = frozenset({"pex"})
 FILE_SKIP_COMMENTS: Tuple[str, ...] = (
     "isort:" + "skip_file",
@@ -68,6 +88,7 @@ DEFAULT_SKIP: FrozenSet[str] = frozenset(
         ".pants.d",
         ".direnv",
         "node_modules",
+        "__pypackages__",
     }
 )
 
@@ -213,8 +234,12 @@ class _Config:
     reverse_sort: bool = False
     star_first: bool = False
     import_dependencies = Dict[str, str]
+    git_ignore: Dict[Path, Set[Path]] = field(default_factory=dict)
+    format_error: str = "{error}: {message}"
+    format_success: str = "{success}: {message}"
+    sort_order: str = "natural"
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         py_version = self.py_version
         if py_version == "auto":  # pragma: no cover
             if sys.version_info.major == 2 and sys.version_info.minor <= 6:
@@ -255,7 +280,7 @@ class _Config:
                 f"{self.wrap_length} > {self.line_length}."
             )
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return id(self)
 
 
@@ -268,12 +293,13 @@ class Config(_Config):
         settings_file: str = "",
         settings_path: str = "",
         config: Optional[_Config] = None,
-        **config_overrides,
+        **config_overrides: Any,
     ):
         self._known_patterns: Optional[List[Tuple[Pattern[str], str]]] = None
         self._section_comments: Optional[Tuple[str, ...]] = None
         self._skips: Optional[FrozenSet[str]] = None
         self._skip_globs: Optional[FrozenSet[str]] = None
+        self._sorting_function: Optional[Callable[..., List[str]]] = None
 
         if config:
             config_vars = vars(config).copy()
@@ -283,6 +309,7 @@ class Config(_Config):
             config_vars.pop("_section_comments")
             config_vars.pop("_skips")
             config_vars.pop("_skip_globs")
+            config_vars.pop("_sorting_function")
             super().__init__(**config_vars)  # type: ignore
             return
 
@@ -422,11 +449,13 @@ class Config(_Config):
         if "src_paths" not in combined_config:
             combined_config["src_paths"] = (path_root / "src", path_root)
         else:
-            src_paths: List[Path] = []
+            src_paths: Set[Path] = set()
             for src_path in combined_config.get("src_paths", ()):
-                full_path = path_root / src_path
-                if full_path not in src_paths:
-                    src_paths.append(full_path)
+                full_paths = (
+                    path_root.glob(src_path) if "*" in str(src_path) else [path_root / src_path]
+                )
+                for path in full_paths:
+                    src_paths.add(path)
 
             combined_config["src_paths"] = tuple(src_paths)
 
@@ -456,7 +485,8 @@ class Config(_Config):
                 warn(
                     "W0503: Deprecated config options were used: "
                     f"{', '.join(deprecated_options_used)}."
-                    "Please see the 5.0.0 upgrade guide: bit.ly/isortv5."
+                    "Please see the 5.0.0 upgrade guide: "
+                    "https://pycqa.github.io/isort/docs/upgrade_guides/5.0.0.html"
                 )
 
         if known_other:
@@ -481,7 +511,7 @@ class Config(_Config):
 
         super().__init__(sources=tuple(sources), **combined_config)  # type: ignore
 
-    def is_supported_filetype(self, file_name: str):
+    def is_supported_filetype(self, file_name: str) -> bool:
         _root, ext = os.path.splitext(file_name)
         ext = ext.lstrip(".")
         if ext in self.supported_extensions:
@@ -507,6 +537,39 @@ class Config(_Config):
         else:
             return bool(_SHEBANG_RE.match(line))
 
+    def _check_folder_gitignore(self, folder: str) -> Optional[Path]:
+        env = {"LANG": "C.UTF-8"}
+        try:
+            topfolder_result = subprocess.check_output(  # nosec # skipcq: PYL-W1510
+                ["git", "-C", folder, "rev-parse", "--show-toplevel"], encoding="utf-8", env=env
+            )
+        except subprocess.CalledProcessError:
+            return None
+
+        git_folder = Path(topfolder_result.rstrip()).resolve()
+
+        files: List[str] = []
+        # don't check symlinks; either part of the repo and would be checked
+        # twice, or is external to the repo and git won't know anything about it
+        for root, _dirs, git_files in os.walk(git_folder, followlinks=False):
+            if ".git" in _dirs:
+                _dirs.remove(".git")
+            for git_file in git_files:
+                files.append(os.path.join(root, git_file))
+        git_options = ["-C", str(git_folder), "-c", "core.quotePath="]
+        try:
+            ignored = subprocess.check_output(  # nosec # skipcq: PYL-W1510
+                ["git", *git_options, "check-ignore", "-z", "--stdin", "--no-index"],
+                encoding="utf-8",
+                env=env,
+                input="\0".join(files),
+            )
+        except subprocess.CalledProcessError:
+            return None
+
+        self.git_ignore[git_folder] = {Path(f) for f in ignored.rstrip("\0").split("\0")}
+        return git_folder
+
     def is_skipped(self, file_path: Path) -> bool:
         """Returns True if the file and/or folder should be skipped based on current settings."""
         if self.directory and Path(self.directory) in file_path.resolve().parents:
@@ -516,16 +579,6 @@ class Config(_Config):
 
         os_path = str(file_path)
 
-        if self.skip_gitignore:
-            if file_path.name == ".git":  # pragma: no cover
-                return True
-
-            result = subprocess.run(  # nosec # skipcq: PYL-W1510
-                ["git", "-C", str(file_path.parent), "check-ignore", "--quiet", os_path]
-            )
-            if result.returncode == 0:
-                return True
-
         normalized_path = os_path.replace("\\", "/")
         if normalized_path[1:2] == ":":
             normalized_path = normalized_path[2:]
@@ -542,17 +595,34 @@ class Config(_Config):
                 return True
             position = os.path.split(position[0])
 
-        for glob in self.skip_globs:
-            if fnmatch.fnmatch(file_name, glob) or fnmatch.fnmatch("/" + file_name, glob):
+        for sglob in self.skip_globs:
+            if fnmatch.fnmatch(file_name, sglob) or fnmatch.fnmatch("/" + file_name, sglob):
                 return True
 
         if not (os.path.isfile(os_path) or os.path.isdir(os_path) or os.path.islink(os_path)):
             return True
 
+        if self.skip_gitignore:
+            if file_path.name == ".git":  # pragma: no cover
+                return True
+
+            git_folder = None
+
+            file_paths = [file_path, file_path.resolve()]
+            for folder in self.git_ignore:
+                if any(folder in path.parents for path in file_paths):
+                    git_folder = folder
+                    break
+            else:
+                git_folder = self._check_folder_gitignore(str(file_path.parent))
+
+            if git_folder and any(path in self.git_ignore[git_folder] for path in file_paths):
+                return True
+
         return False
 
     @property
-    def known_patterns(self):
+    def known_patterns(self) -> List[Tuple[Pattern[str], str]]:
         if self._known_patterns is not None:
             return self._known_patterns
 
@@ -599,6 +669,29 @@ class Config(_Config):
         self._skip_globs = self.skip_glob.union(self.extend_skip_glob)
         return self._skip_globs
 
+    @property
+    def sorting_function(self) -> Callable[..., List[str]]:
+        if self._sorting_function is not None:
+            return self._sorting_function
+
+        if self.sort_order == "natural":
+            self._sorting_function = sorting.naturally
+        elif self.sort_order == "native":
+            self._sorting_function = sorted
+        else:
+            available_sort_orders = ["natural", "native"]
+            import pkg_resources
+
+            for sort_plugin in pkg_resources.iter_entry_points("isort.sort_function"):
+                available_sort_orders.append(sort_plugin.name)
+                if sort_plugin.name == self.sort_order:
+                    self._sorting_function = sort_plugin.load()
+                    break
+            else:
+                raise SortingFunctionDoesNotExist(self.sort_order, available_sort_orders)
+
+        return self._sorting_function
+
     def _parse_known_pattern(self, pattern: str) -> List[str]:
         """Expand pattern if identified as a directory and return found sub packages"""
         if pattern.endswith(os.path.sep):
@@ -613,8 +706,10 @@ class Config(_Config):
         return patterns
 
 
-def _get_str_to_type_converter(setting_name: str) -> Callable[[str], Any]:
-    type_converter: Callable[[str], Any] = type(_DEFAULT_SETTINGS.get(setting_name, ""))
+def _get_str_to_type_converter(setting_name: str) -> Union[Callable[[str], Any], Type[Any]]:
+    type_converter: Union[Callable[[str], Any], Type[Any]] = type(
+        _DEFAULT_SETTINGS.get(setting_name, "")
+    )
     if type_converter == WrapModes:
         type_converter = wrap_mode_from_string
     return type_converter
@@ -704,7 +799,7 @@ def _get_config_data(file_path: str, sections: Tuple[str]) -> Dict[str, Any]:
                             and config_key.endswith("}")
                             and extension
                             in map(
-                                lambda text: text.strip(), config_key[len("*.{") : -1].split(",")
+                                lambda text: text.strip(), config_key[len("*.{") : -1].split(",")  # type: ignore # noqa
                             )
                         ):
                             settings.update(config.items(config_key))
